#!/usr/bin/env python

# simple interface to the available scripts
from argparse import ArgumentParser, RawTextHelpFormatter, SUPPRESS
import sys
from pmx import __version__


class PmxCli:

    def __init__(self):
        parser = ArgumentParser(
            description='''
    ------------------------
    pmx command line scripts
    ------------------------

    Available commands are:
        mutate        Mutate protein or DNA/RNA
        gentop        Fill hybrid topology with B states
        analyse       Estimate free energy from Gromacs xvg files

        atomMapping   Ligand alchemy: map atoms for morphing
        ligandHybrid  Ligand alchemy: hybrid structure/topology

        doublebox     Place two input structures into a single box
        abfe          Setup files for an ABFE calculation

        genlib        Generate pmx ff library
        gmxlib        Show/set GMXLIB path''',
            formatter_class=RawTextHelpFormatter)

        parser.add_argument('-v', '--version', action='version',
                            version=__version__)
        parser.add_argument('command', help=SUPPRESS)
        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print('Unrecognized command')
            parser.print_help()
            exit(1)
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    def mutate(self):
        from . import mutate
        mutate.entry_point()

    def gentop(self):
        from . import generate_hybrid_topology
        generate_hybrid_topology.entry_point()

    def analyse(self):
        from . import analyze_dhdl
        analyze_dhdl.entry_point()

    def atomMapping(self):
        from . import atomMapping
        atomMapping.entry_point()

    def ligandHybrid(self):
        from . import ligandHybrid
        ligandHybrid.entry_point()

    def doublebox(self):
        from . import make_double_box
        make_double_box.entry_point()

    def abfe(self):
        from . import setup_abfe
        setup_abfe.entry_point()

    def genlib(self):
        from . import generate_hybrid_residue
        generate_hybrid_residue.entry_point()

    def gmxlib(self):
        from . import set_gmxlib
        set_gmxlib.entry_point()


def check_unknown_cmd(unknowns):
    '''Checks unknown command line arguments are raises a warning if unexpected
    commands are found.
    '''
    expected = ['pmx', 'analyse', 'mutate', 'doublebox', 'gentop', 'gmxlib',
                'genlib', 'abfe', 'atomMapping', 'ligandHybrid']

    for cmd in unknowns:
        if cmd not in expected:
            print('Unknown command found in your command line: "{}". '
                  'This command will be ignored'.format(cmd))


def entry_point():
    PmxCli()


if __name__ == '__main__':
    entry_point()



#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""docs to be added
"""

import sys
from .library import _aliases, pmx_data_file
from .parser import parseList, kickOutComments, readSection


class RTPParser:

    def __init__(self, filename=None):
        self.keys = []
        self.entries = {}
        self.lines = []
        self.filename = ''
        self.__cur_id = 0
        if filename is not None:
            self.parse(filename)

    def parse(self, filename):
        filename = pmx_data_file(filename)
        self.filename = filename
        l = open(filename).readlines()
        self.lines = kickOutComments(l, ';')
        self.__get_residue_names()
        for key in self.keys:
            rtp_lines = self.__read_residue_entry(key)
            # read atoms
            al = readSection(rtp_lines, '[ atoms ]', '[')
            atoms = self.__read_rtp_atoms(key, al)
            # read bonds
            bl = readSection(rtp_lines, '[ bonds ]', '[')
            bonds = self.__read_rtp_bonds(key, bl)
            # read dihedrals
            dl = readSection(rtp_lines, '[ dihedrals ]', '[')
            diheds = self.__read_rtp_dihedrals(key, dl)
            # read impropers
            il = readSection(rtp_lines, '[ impropers ]', '[')
            improps = self.__read_rtp_impropers(key, il)
            # read cmap (only for charmm)
            cmap = readSection(rtp_lines, '[ cmap ]', '[')

            self.entries[key] = {
                'atoms': atoms,
                'bonds': bonds,
                'diheds': diheds,
                'improps': improps,
                'cmap': cmap
                }

    def __str__(self):
        s = '<%s | "%s" |  %d residue entries >' % (self.__class__,
                                                    self.filename,
                                                    len(self.keys))
        return s

    def __getitem__(self, item):
        return self.entries[item]

    def __delitem__(self, item):
        del self.entries[item]
        self.keys.remove(item)

    def add_entry(self, name, rtp_dic):
        if name in self.keys:
            del self.entries[name]
        else:
            self.keys.append(name)
        self.entries[name] = rtp_dic

    def __contains__(self, key):
        return key in self.keys

    def __iter__(self):
        return self

    def next(self):
        if self.__cur_id >= len(self.keys):
            self.__cur_id = 0
            raise StopIteration
        name = self.keys[self.__cur_id]
        entr = self.entries[self.keys[self.__cur_id]]
        self.__cur_id += 1
        return name, entr

    def write(self, out_file):
        if not hasattr(out_file, "write"):
            fp = open(out_file, "w")
        else:
            fp = out_file
        for key in self.keys:
            entr = self.entries[key]
            print('[ %s ]' % key, file=fp)
            print(' [ atoms ]', file=fp)
            for atom in entr['atoms']:
                print("%6s   %-15s  %8.5f  %d" % (atom[0], atom[1], atom[2], atom[3]), file=fp)
            if entr['bonds']:
                print(' [ bonds ]', file=fp)
                for bond in entr['bonds']:
                    print("%6s  %6s" % (bond[0], bond[1]), file=fp)
            if entr['diheds']:
                print(' [ dihedrals ]', file=fp)
                for dih in entr['diheds']:
                    print("%6s  %6s  %6s  %6s  %-25s" % (dih[0], dih[1], dih[2], dih[3], dih[4]), file=fp)
            if entr['improps']:
                print(' [ impropers ]', file=fp)
                for dih in entr['improps']:
                    try:
                        print("%6s  %6s  %6s  %6s  %-25s" % (dih[0], dih[1], dih[2], dih[3], dih[4]), file=fp)
                    except:
                        print("%6s  %6s  %6s  %6s " % (dih[0], dih[1], dih[2], dih[3]), file=fp)
            print('', file=fp)

    def __check_residue_tree(self, model):
        for c in model.chains:
            if not c.residue_tree_ok:
                print('pmx_Error_> Broken residue tree in chain ', c.id, file=sys.stderr)
                sys.exit(1)

    def assign_params(self, model):
        self.__check_residue_tree(model)
        self.__assign_atom_params(model)
        self.__make_bonds(model)
        self.__make_angles(model)
        self.__make_dihedrals(model)
        self.__make_impropers(model)

    def assign_dihedral_params(self, model, directives):
        for dih in model.dihedral_list:
            a1, a2, a3, a4 = dih[:4]
            resid = a2.resnr  # FIXME: variable not used
            rtp_d = self.entries[a2.resname]  # FIXME: variable not used
            name2 = a2.name
            if a1.resnr == a2.resnr:
                name1 = a1.name
            elif a1.resnr == a2.resnr - 1:
                name1 = '-'+a1.name
            elif a1.resnr == a2.resnr + 1:
                name1 = '+'+a1.name
            if a3.resnr == a2.resnr:
                name3 = a3.name
            elif a3.resnr == a2.resnr - 1:
                name3 = '-'+a3.name
            elif a3.resnr == a2.resnr + 1:
                name3 = '+'+a3.name
            if a4.resnr == a2.resnr:
                name4 = a4.name
            elif a4.resnr == a2.resnr - 1:
                name4 = '-'+a4.name
            elif a1.resnr == a2.resnr + 1:
                name4 = '+'+a4.name
            d = self.__find_rtp_dihedral(a2.resname, name1, name2, name3, name4)
            if d is not None:
                if d[4] in directives:
                    dih = dih[:5]+directives[d[4]]
                else:
                    raise ValueError('No directive found')

    def __assign_atom_params(self, model):
        for residue in model.residues:
            rtp_atoms = self.entries[residue.resname]['atoms']
            for atom_entry in rtp_atoms:
                name = atom_entry[0]
                atom = residue[name]
                atom.atomtype = atom_entry[1]
                atom.q = atom_entry[2]
                atom.cgnr = atom_entry[3]

    def __make_bonds(self, model):
        model.bond_list = []
        for residue in model.residues:
            rtp_bonds = self.entries[residue.resname]['bonds']
            for a1, a2 in rtp_bonds:
                if a1[0] not in ['+', '-'] and a2[0] not in ['+', '-']:
                    atom1 = residue[a1]
                    atom2 = residue[a2]
                    atom1.bonds.append(atom2)
                    atom2.bonds.append(atom1)
                    model.bond_list.append([atom1, atom2])
                else:
                    if a1[0] == '-':
                        atom1 = residue.previous[a1[1:]]
                    elif a1[0] == '+':
                        atom1 = residue.next[a1[1:]]
                    else:
                        atom1 = residue[a1]
                    if a2[0] == '-':
                        atom2 = residue.previous[a2[1:]]
                    elif a2[0] == '+':
                        atom2 = residue.next[a2[1:]]
                    else:
                        atom2 = residue[a2]
                    if atom2 not in atom1.bonds:
                        atom1.bonds.append(atom2)
                        atom2.bonds.append(atom1)
                        model.bond_list.append([atom1, atom2])

        # cys
        for residue in model.residues:
            if residue.resname in ['CYS2', 'CCYX']:
                for r in model.residues:
                    if r.resname in ['CYS2', 'CCYX'] and r.id < residue.id:
                        sg1 = residue['SG']
                        sg2 = r['SG']
                        d = sg1 - sg2
                        if d < 2.5:
                            print('pmx__> Disulfid bond between residue', sg1.resnr, 'and', sg2.resnr, file=sys.stderr)
                            sg1.bonds.append(sg2)
                            sg2.bonds.append(sg1)
                            model.bond_list.append([sg1, sg2])

        for atom in model.atoms:
            for b in atom.bonds:
                atom.connected.append(b)

    def __make_angles(self, model):
        model.angle_list = []
        for atom in model.atoms:
            for b in atom.bonds:
                for bb in b.bonds:
                    if bb.id < atom.id:
                        atom.b13.append(bb)
                        bb.b13.append(atom)
                        atom.connected.append(bb)
                        bb.connected.append(atom)
                        model.angle_list.append([atom, b, bb])

    def __make_dihedrals(self, model):
        model.dihedral_list = []
        for atom in model.atoms:
            for b in atom.bonds:
                for bb in b.bonds:
                    if bb.id != atom.id:
                        for bbb in bb.bonds:
                            if atom.id < bbb.id:
                                if bbb not in atom.bonds:
                                    model.dihedral_list.append([atom, b, bb, bbb])
                                    if bbb not in atom.connected:
                                        atom.b14.append(bbb)
                                        bbb.b14.append(atom)
                                        atom.connected.append(bbb)
                                        bbb.connected.append(atom)

    def __make_impropers(self, model):
        model.improper_list = []
        for residue in model.residues:
            rtp_imp = self.entries[residue.resname]['improps']

            for imp in rtp_imp:
                atoms = []
                for name in imp[:4]:
                    if name[0] == '+':
                        atom = residue.next[name[1:]]
                    elif name[0] == '-':
                        atom = residue.previous[name[1:]]
                    else:
                        atom = residue[name]
                    atoms.append(atom)
                assert len(atoms) == 4
                model.improper_list.append(atoms + [imp[4]])

    def __get_residue_names(self):
        self.keys = []
        for line in self.lines:
            if line.strip().startswith('['):
                if line.strip()[1:-1].strip() not in \
                       ['atoms', 'bonds', 'dihedrals', 'impropers',
                        'bondedtypes']:
                    self.keys.append(line.strip()[1:-1].strip())

    def __read_residue_entry(self, key):
        r = []
        for line in self.lines:
            if line.strip()[1:-1].strip() == key:
                idx = self.lines.index(line)
        for line in self.lines[idx+1:]:
            if line.strip().startswith('['):
                if line.strip()[1:-1].strip() not in \
                       ['atoms', 'bonds', 'dihedrals', 'impropers', 'cmap']:
                    break
                else:
                    r.append(line)
            else:
                r.append(line)
        return r

    def __read_rtp_atoms(self, resname, lines):
        atoms = []
        for line in lines:
            entr = line.split()
            if resname in _aliases:
                if entr[0] in _aliases[resname]:
                    entr[0] = _aliases[resname][entr[0]]
            entr[2] = float(entr[2])
            entr[3] = int(entr[3])
            atoms.append(entr)
        return atoms

    def __read_rtp_bonds(self, resname, lines):
        bonds = []
        for line in lines:
            entr = line.split()
            if entr[0] not in ['+', '-']:
                if resname in _aliases:
                    if entr[0] in _aliases[resname]:
                        entr[0] = _aliases[resname][entr[0]]
            if entr[1] not in ['+', '-']:
                if resname in _aliases:
                    if entr[1] in _aliases[resname]:
                        entr[1] = _aliases[resname][entr[1]]
            bonds.append(entr)
        return bonds

    def __read_rtp_dihedrals(self, resname, lines):
        diheds = []
        for line in lines:
            entr = line.split()
            if entr[0] not in ['+', '-']:
                if resname in _aliases:
                    if entr[0] in _aliases[resname]:
                        entr[0] = _aliases[resname][entr[0]]
            if entr[1] not in ['+', '-']:
                if resname in _aliases:
                    if entr[1] in _aliases[resname]:
                        entr[1] = _aliases[resname][entr[1]]
            if entr[2] not in ['+', '-']:
                if resname in _aliases:
                    if entr[2] in _aliases[resname]:
                        entr[2] = _aliases[resname][entr[2]]
            if entr[3] not in ['+', '-']:
                if resname in _aliases:
                    if entr[3] in _aliases[resname]:
                        entr[3] = _aliases[resname][entr[3]]
            if len(entr) == 5:
                diheds.append(entr)
            else:
                diheds.append(entr+[''])
        return diheds

    def __read_rtp_impropers(self, resname, lines):
        improps = []
        for line in lines:
            entr = line.split()
            if entr[0] not in ['+', '-']:
                if resname in _aliases:
                    if entr[0] in _aliases[resname]:
                        entr[0] = _aliases[resname][entr[0]]
            if entr[1] not in ['+', '-']:
                if resname in _aliases:
                    if entr[1] in _aliases[resname]:
                        entr[1] = _aliases[resname][entr[1]]
            if entr[2] not in ['+', '-']:
                if resname in _aliases:
                    if entr[2] in _aliases[resname]:
                        entr[2] = _aliases[resname][entr[2]]
            if entr[3] not in ['+', '-']:
                if resname in _aliases:
                    if entr[3] in _aliases[resname]:
                        entr[3] = _aliases[resname][entr[3]]
            if len(entr) == 5:
                improps.append(entr)
            else:
                improps.append(entr+[''])
        return improps

    def __find_rtp_dihedral(self, key, name1, name2, name3, name4):
        rtp_d = self.entries[key]['diheds']
        for d in rtp_d:
            if (d[0] == name1 and d[1] == name2 and
                d[2] == name3 and d[3] == name4) or \
                (d[0] == name4 and d[1] == name3 and
                 d[2] == name2 and d[3] == name1):
                return d
        return None

    def __find_rtp_improper(self, key, name1, name2, name3, name4):
        rtp_i = self.entries[key]['improps']
        for d in rtp_i:
            if (d[0] == name1 and d[1] == name2 and
                d[2] == name3 and d[3] == name4) or \
                (d[0] == name4 and d[1] == name3 and
                 d[2] == name2 and d[3] == name1):
                return d
        return None


class BondedParser:

    def __init__(self, filename=None, version='old'):
        self.lines = []
        self.filename = ''
        self.directives = {}
        self.bondtypes = []
        self.angletypes = []
        self.dihedraltypes = []
        if filename:
            self.parse(filename)

    def parse(self, filename):
        if not hasattr(filename, "append"):  # not a list
            filename = pmx_data_file(filename)
            self.filename = filename
            l = open(filename).readlines()
        else:
            l = filename
            self.filename = '< from list >'
        self.lines = kickOutComments(l, ';')
        self.__parse_directives()
        self.__parse_bondtypes()
        self.__parse_angletypes()
        self.__parse_dihedraltypes()

    def __str__(self):
        s = '< %s | %s >' % (self.__class__, self.filename)
        return s

    def assign_params(self, model):
        for bond in model.bond_list:
            params = self.get_bond_param(bond[0].bond_type, bond[1].bond_type)
            bond.extend(params)
        for angle in model.angle_list:
            params = self.get_angle_param(angle[0].bond_type,
                                          angle[1].bond_type,
                                          angle[2].bond_type)
            angle.extend(params)
        for dih in model.dihedral_list:
            params = self.get_dihedral_param(dih[0].bond_type,
                                             dih[1].bond_type,
                                             dih[2].bond_type,
                                             dih[3].bond_type,
                                             3)
            dih.extend(params)

        for dih in model.improper_list:
            if dih[4] == '':
                params = self.get_dihedral_param(dih[0].bond_type,
                                                 dih[1].bond_type,
                                                 dih[2].bond_type,
                                                 dih[3].bond_type,
                                                 1)
                dih.extend(params[1:])
            else:
                dih = dih[:4] + [1] + self.directives[dih[4]]

    def get_bond_param(self, type1, type2):
        for entr in self.bondtypes:
            if (type1 == entr[0] and type2 == entr[1]) or \
               (type1 == entr[1] and type2 == entr[0]):
                return entr[2:]
        return None

    def get_angle_param(self, type1, type2, type3):
        for entr in self.angletypes:
            if (type1 == entr[0] and type2 == entr[1] and type3 == entr[2]) or \
               (type1 == entr[2] and type2 == entr[1] and type3 == entr[0]):
                return entr[3:]
        return None

    def get_dihedral_param(self, type1, type2, type3, type4, func):

        self.result = []
        found = 0
        for entr in self.dihedraltypes:
            if (type1 == entr[0] and
                type2 == entr[1] and
                type3 == entr[2] and
                type4 == entr[3] and func == entr[4]) or \
                (type1 == entr[3] and
                 type2 == entr[2] and
                 type3 == entr[1] and
                 type4 == entr[0] and func == entr[4]):
                self.result.append(entr[4:])
                found = 1
                if(func != 9):
                    return self.result

        if found == 1:
            return self.result

        for entr in self.dihedraltypes:
            if ('X' == entr[0] and
                type2 == entr[1] and
                type3 == entr[2] and
                type4 == entr[3] and func == entr[4]) or \
                (type1 == entr[0] and
                 type2 == entr[1] and
                 type3 == entr[2] and
                 'X' == entr[3] and func == entr[4]):
                self.result.append(entr[4:])
                if(func != 9):
                    return self.result
            if ('X' == entr[3] and
                type2 == entr[2] and
                type3 == entr[1] and
                type4 == entr[0] and func == entr[4]) or \
                (type1 == entr[3] and
                 type2 == entr[2] and
                 type3 == entr[1] and
                 'X' == entr[0] and func == entr[4]):
                self.result.append(entr[4:])
                found = 1
                if(func != 9):
                    return self.result

        if found == 1:
            return self.result

        for entr in self.dihedraltypes:
            if ('X' == entr[0] and
                type2 == entr[1] and
                type3 == entr[2] and
                'X' == entr[3] and func == entr[4]) or \
                ('X' == entr[3] and
                 type2 == entr[2] and
                 type3 == entr[1] and
                 'X' == entr[0] and func == entr[4]):
                self.result.append(entr[4:])
                found = 1
                if(func != 9):
                    return self.result

        if found == 1:
            return self.result

        for entr in self.dihedraltypes:
            if ('X' == entr[0] and
                'X' == entr[1] and
                type3 == entr[2] and
                type4 == entr[3] and func == entr[4]) or \
                (type1 == entr[3] and
                 type2 == entr[2] and
                 'X' == entr[1] and
                 'X' == entr[0] and func == entr[4]):
                self.result.append(entr[4:])
                found = 1
                if(func != 9):
                    return self.result

        if found == 1:
            return self.result

        for entr in self.dihedraltypes:  # exchange 1 -> 3
            if ('X' == entr[0] and
                'X' == entr[1] and
                type3 == entr[2] and
                type2 == entr[3] and func == entr[4]) or \
                (type1 == entr[3] and
                 type4 == entr[2] and
                 'X' == entr[1] and
                 'X' == entr[0] and func == entr[4]):
                self.result.append(entr[4:])
                if(func != 9):
                    return self.result

        for entr in self.dihedraltypes:
            if (type1 == entr[0] and
                'X' == entr[1] and
                'X' == entr[2] and
                type4 == entr[3] and func == entr[4]) or \
                (type1 == entr[3] and
                 'X' == entr[1] and
                 'X' == entr[2] and
                 type4 == entr[0] and func == entr[4]):
                self.result.append(entr[4:])
                if(func != 9):
                    return self.result

        return self.result

    def __parse_directives(self):
        for line in self.lines:
            if line.startswith('#define'):
                entr = line.split()
                name = entr[1]
                params = [float(x) for x in entr[2:]]
                self.directives[name] = params
        self.lines = kickOutComments(self.lines, '#')

    def __parse_bondtypes(self):
        res = []
        starts = []
        for i, line in enumerate(self.lines):
            if line.strip().startswith('[ bondtypes ]'):
                starts.append(i)
        for s in starts:
            lst = readSection(self.lines[s:], '[ bondtypes ]', '[')
            lst = parseList('ssiff', lst)
            res.extend(lst)
        self.bondtypes = res

    def __parse_angletypes(self):
        res = []
        starts = []
        for i, line in enumerate(self.lines):
            if line.strip().startswith('[ angletypes ]'):
                starts.append(i)
        for s in starts:
            lst = readSection(self.lines[s:], '[ angletypes ]', '[')
            try:
                lst = parseList('sssiff', lst)
            except:
                try:
                    lst = parseList('sssiffff', lst)
                except:
                    raise ValueError("Unkown Angle type")
            res.extend(lst)
        self.angletypes = res

    def __parse_dihedraltypes(self):
        res = []
        starts = []
        for i, line in enumerate(self.lines):
            if line.strip().startswith('[ dihedraltypes ]'):
                starts.append(i)
        for s in starts:
            lst = readSection(self.lines[s:], '[ dihedraltypes ]', '[')
            try:
                lst = parseList('ssssiffffff', lst)
            except:
                try:
                    lst = parseList('ssssiffi', lst)
                except:
                    try:
                        lst = parseList('ssiffi', lst)
                    except:
                        lst = parseList('ssssiff', lst)
            res.extend(lst)
        self.dihedraltypes = res


class NBParser:

    def __init__(self, filename=None, version='new', ff='amber'):
        self.lines = []
        self.ff = ff
        self.filename = ''
        self.atomtypes = {}
        if filename is not None:
            self.parse(filename, version)

    def parse(self, filename, version):
        if not hasattr(filename, "append"):  # not a list
            filename = pmx_data_file(filename)
            l = open(filename).readlines()
            self.filename = filename
        else:
            l = filename
            self.filename = '< from list >'
        self.lines = kickOutComments(l, ';')
        self.__parse_atomtypes(version)

    def __str__(self):
        s = '< %s | %s >' % (self.__class__, self.filename)
        return s

    def __parse_atomtypes(self, version):

        self.atomtypes = {}
        lst = readSection(self.lines, '[ atomtypes ]', '[')
        ffnamelower = self.ff.lower()
        if version == 'old':
            if ffnamelower.startswith('amber'):
                lst = parseList('ssffsff', lst)
                for entr in lst:
                    self.atomtypes[entr[0]] = {
                        'bond_type': entr[1],
                        'mass': float(entr[2]),
                        'sigma': entr[5]*10,  # nm -> A
                        'eps': entr[6]
                        }
            elif ffnamelower.startswith('opls'):
                lst = parseList('ssiffsff', lst)
                for entr in lst:
                    self.atomtypes[entr[0]] = {
                        'bond_type': entr[1],
                        'mass': float(entr[3]),
                        'sigma': entr[6]*10,  # nm -> A
                        'eps': entr[7]
                        }

        elif version == 'new':
            if ffnamelower.startswith('amber') or ffnamelower.startswith('charmm'):
                lst = parseList('siffsff', lst)
                for entr in lst:
                    self.atomtypes[entr[0]] = {
                        'bond_type': entr[0],
                        'mass': float(entr[2]),
                        'sigma': entr[5]*10,  # nm -> A
                        'eps': entr[6]
                        }
            elif ffnamelower.startswith('opls'):
                lst = parseList('ssiffsff', lst)
                for entr in lst:
                    self.atomtypes[entr[0]] = {
                        'bond_type': entr[1],
                        'mass': float(entr[3]),
                        'sigma': entr[6]*10,  # nm -> A
                        'eps': entr[7]
                        }
            elif ffnamelower.startswith('gaff'):
                try:
                    lst = parseList('sffsff', lst)
                    for entr in lst:
                        self.atomtypes[entr[0]] = {
                            'bond_type': entr[1],
                            'mass': float(entr[2]),
                            # 'charge':float(entr[3]),
                            'sigma': entr[4]*10,  # nm -> A
                            'eps': entr[5]
                            }
                except:
                    lst = parseList('ssffsff', lst)
                    for entr in lst:
                        self.atomtypes[entr[0]] = {
                            'bond_type': entr[1],
                            'mass': float(entr[2]),
                            'charge': float(entr[3]),
                            'sigma': entr[5]*10,  # nm -> A
                            'eps': entr[6]
                            }

    def assign_params(self, model):
        for atom in model.atoms:
            atom.bond_type = self.atomtypes[atom.atomtype]['bond_type']
            atom.sigma = self.atomtypes[atom.atomtype]['sigma']
            atom.eps = self.atomtypes[atom.atomtype]['eps']


class ATPParser:

    def __init__(self,  fn='ffamber99sb.atp'):
        self.fn = fn
        self.dic = {}
        if fn is not None:
            self.parse()

    def parse(self):
        lst = open(self.fn).readlines()
        lst = kickOutComments(lst, ';')
        lst = parseList('sf', lst)
        for tp,  mass in lst:
            self.dic[tp] = mass

    def __getitem__(self,  item):
        return self.dic[item]



import sys, os
import copy as cp
import numpy as np
from pmx import *
from pmx import geometry
from pmx import ndx
from pmx import library
from pmx.options import *
#from pmx.forcefield2 import *
from AbsRestraints import *

def main(argv):

    desc=('Read ligand and protein structure and topologies.',
          'Add distance, angle and dihedral restraints.',
          '',)

# define input/output files

    files= [
        FileOption("-f", "r",["pdb","gro"],"system.pdb", "input system structure"),
        FileOption("-ilig", "r/o",["itp","top"],"", "input ligand topology"),
        FileOption("-iprot", "r/o",["itp","top"],"", "input protein topology"),
#        FileOption("-fprot", "r",["pdb","gro"],"protein.pdb", "input protein structure"),
        FileOption("-n", "r",["ndx"],"ndx_restr.ndx", "input index file"),
#        FileOption("-oitp", "w",["itp"],"out.itp", "output renumbered ligand.itp"),
        FileOption("-oii", "w",["itp"],"outii.itp", "intermolecular_interactions itp to be added to the end of .top"),
        FileOption("-dg", "w",["dat"],"out_dg.dat", "write contribution of the restraints to the free energy"),
        FileOption("-orestr", "w/o",["pdb"],"restrained_atoms.pdb", "optional output of restrained atoms only"),
        ]

# define options

    options=[
       Option( "-manLig", "bool", False, "manual selection of ligand atoms for restraints"),
       Option( "-manProt", "bool", False, "manual selection of protein atoms for restraints"),
       Option( "-ligCenter", "string", False, "manually provide only central ligand atom"),
       Option( "-ligMaxSpread", "bool", False, "for automated ligand atom detection spread atoms to maximize distance between them"),
       Option( "-T", "float", 300.0, "temperature"),
       Option( "-kBond", "float", 4184.0, "force constant for bond, kJ/mol"),
       Option( "-kAngle", "float", 41.84, "force constant for angles, kJ/mol"),
       Option( "-kDih", "float", 41.84, "force constant for dihedrals, kJ/mol"),
       Option( "-martini", "bool", False, "martini force field"),
        ]

    help_text = ('Read ligand+protein structure and topologies.',
          'Add distance, angle and dihedral restraints.',
          '',)

# pass options, files and the command line to pymacs

    cmdl = Commandline( argv, options = options, fileoptions = files, program_desc = help_text, check_for_existing_files = False, version = "0.0" )

#    system = Model(cmdl['-f'])
#    protModel = Model(cmdl['-fprot'])


    absRestr = AbsRestraints( strFile=cmdl['-f'], topLigFile=cmdl['-ilig'], topProtFile=cmdl['-iprot'], ndxFile=cmdl['-n'], outiiFile=cmdl['-oii'], bManLig=cmdl['-manLig'], bManProt=cmdl['-manProt'], T=cmdl['-T'], dgFile=cmdl['-dg'], ligCenterByName=cmdl['-ligCenter'],bLigMaxSpread=cmdl['-ligMaxSpread'], kBond=cmdl['-kBond'], kAngle=cmdl['-kAngle'], kDih=cmdl['-kDih'], bMartini=cmdl['-martini'], pdbRestrAtoms=cmdl['-orestr'] )    


main( sys.argv )




#!/usr/bin/env python
import pytest
from pmx.scripts.analyze_dhdl import entry_point
import os
import numpy as np
from numpy.random import normal,seed
from numpy.testing import assert_almost_equal
import sys


def test_dhdl(tmpdir):
    #make fake dHdl files in a temp directory
    trjlen=51
    dif=3.0
    W_scale=100.0
    dG=100.0
    noise_scale=10.0
    nruns=80
    l=np.linspace(0., 1., num=trjlen, endpoint=True)
    signal=(1/(1+np.exp((-l+0.5)*10))-0.5)*W_scale + dG
    
    np.random.seed(123456) #force reproducible random numbers
    
    fns_f=[]
    fns_b=[]
    for i in range(nruns):
        f=signal + 0.5*dif + np.random.normal(loc=0.0, scale=noise_scale, size=signal.shape)
        b=signal - 0.5*dif + np.random.normal(loc=0.0, scale=noise_scale, size=signal.shape)
        
        dataf=np.transpose(np.vstack((l,f)))
        datab=np.transpose(np.vstack((l,np.flip(b)))) # don't flip l. it's used as time here, not lambda value
        fh_f=tmpdir.join("test_temp_f%d.xvg"%i)
        fh_b=tmpdir.join("test_temp_b%d.xvg"%i)
        np.savetxt(fh_f, dataf, fmt='%16.8f')
        np.savetxt(fh_b, datab, fmt='%16.8f')
        
        fns_f.append(str(fh_f))
        fns_b.append(str(fh_b))
        
    #run analysis on our fake files
    orig_argv = sys.argv
    orig_dir = os.getcwd()
    os.chdir(tmpdir)
    sys.argv = [['analyze_dhdl.py'],\
                ['-fA'], [x for x in fns_f],\
                ['-fB'], [x for x in fns_b]]
    sys.argv = [item for sublist in sys.argv for item in sublist]
    entry_point()
    sys.argv = orig_argv #reset argv
    os.chdir(orig_dir)   #reset cwd

    #read results
    bar=-1
    barerr=-1
    print("#"*40)
    with open(str(tmpdir)+"/results.txt","rb") as fp:
        lines = fp.readlines()
        for line in lines:
            line=line.decode()
            print(line)
            if("BAR: dG =" in line):
                s=line.split()
                print(s)
                bar=float(s[-2])
            elif("BAR: Std Err (analytical) =" in line):
                s=line.split()
                print(s)
                barerr=float(s[-2])
    
    assert_almost_equal(bar, dG*0.98, decimal=2)
    assert_almost_equal(barerr, 0.17, decimal=2)
    


#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""Functions to read the mutation database
"""

from .model import Model
from .atom import Atom
from .molecule import Molecule
from .parser import readSection, kickOutComments, parseList


def read_mutpdb(filename='mutations_oplsaa.pdb'):
    if not hasattr(filename, "read"):
        s = open(filename).read().split('ENDMDL')[:-1]
    else:
        s = filename.read().split('ENDMDL')[:-1]
    ml = []
    for entr in s:
        m = Model(pdbline=entr)
        ml.append(m)
    rdic = {}
    for m in ml:
        r = m.residues[0]
        rdic[r.resname] = r
    return rdic


def read_new_mtp_entry(entry, filename='mutres.mtp'):

    if not hasattr(filename, "read"):
        lst = open(filename).readlines()
    else:
        lst = filename.readlines()

    lst = kickOutComments(lst, ';')
    key = '[ '+entry+' ]'
    keyw = ('[ morphes ]', '[ atoms ]', '[ impropers ]', '[ dihedrals ]',
            '[ rotations ]', '[ coords ]')
    res = []
    for i, line in enumerate(lst):
        if line.startswith(key):
            for line2 in lst[i+1:]:
                if line2.startswith('[') and \
                   line2 not in keyw:
                    break
                else:
                    res.append(line2)

    morphes = {}
    ml = readSection(res, '[ morphes ]', '[')
    for i, line in enumerate(ml):
        entr = line.split()
        n0 = entr[0]
        t0 = entr[1]
        n1 = entr[3]
        t1 = entr[4]
        morphes[n0] = {
            't0': t0,
            'n1': n1,
            't1': t1,
            }
    atoms = []
    al = readSection(res, '[ atoms ]', '[')
    for i, line in enumerate(al):
        entr = line.split()
        name = entr[0]
        atomtype = entr[1]
        q = float(entr[2])
        cgnr = int(entr[3])
        m = float(entr[4])
        atomtypeB = entr[5]
        qB = float(entr[6])
        mB = float(entr[7])
        a = Atom(name=name, id=i+1, atomtype=atomtype, q=q, m=m, cgnr=cgnr,
                 atomtypeB=atomtypeB, qB=qB, mB=mB)
        atoms.append(a)
    mol = Molecule(atoms=atoms, unity='nm')
    mol.set_resname(entry)

    coords = readSection(res, '[ coords ]', '[')
    coords = parseList('fff', coords)
    for i, atom in enumerate(mol.atoms):
        atom.x = coords[i]
        atom.unity = 'A'
    il = readSection(res, '[ impropers ]', '[')
    imps = []
    for line in il:
        imps.append(line.split())
    diheds = []
    dl = readSection(res, '[ dihedrals ]', '[')
    for line in dl:
        diheds.append(line.split())
    rl = readSection(res, '[ rotations ]', '[')
    rots = []
    for line in rl:
        rots.append(line.split())
    rotdic = {}
    for rot in rots:
        key = rot[0]
        rotdic[key] = rot[1:]

    mol.morphes = morphes
    bonds = []
    return mol, bonds, imps, diheds, rotdic


def read_mtp_entry(entry, filename='ffamber99sb.mtp', version='old'):
    if version == 'new':
        return read_new_mtp_entry(entry, filename=filename)
    if not hasattr(filename, "read"):
        lst = open(filename).readlines()
    else:
        lst = filename.readlines()

    lst = kickOutComments(lst, ';')
    key = '[ '+entry+' ]'
    keyw = ('[ morphes ]', '[ atoms ]', '[ bonds ]', '[ impropers ]',
            '[ dihedrals ]', '[ rotations ]', '[ coords ]')
    res = []
    for i, line in enumerate(lst):
        if line.startswith(key):
            for line2 in lst[i+1:]:
                if line2.startswith('[') and \
                   line2 not in keyw:
                    break
                else:
                    res.append(line2)

    morphes = {}
    ml = readSection(res, '[ morphes ]', '[')
    for i, line in enumerate(ml):
        entr = line.split()
        n0 = entr[0]
        r0 = entr[1]
        t0 = entr[2]
        n1 = entr[4]
        r1 = entr[5]
        t1 = entr[6]
        morphes[n0] = {
            'r0': r0,
            't0': t0,
            'n1': n1,
            'r1': r1,
            't1': t1,
            }

    atoms = []
    al = readSection(res, '[ atoms ]', '[')
    for i, line in enumerate(al):
        entr = line.split()
        name = entr[0]
        atomtype = entr[1]
        q = float(entr[2])
        cgnr = int(entr[3])
        m = float(entr[4])
        atomtypeB = entr[5]
        qB = float(entr[6])
        mB = float(entr[7])
        a = Atom(name=name, id=i+1, atomtype=atomtype, q=q, m=m, cgnr=cgnr,
                 atomtypeB=atomtypeB, qB=qB, mB=mB)
        atoms.append(a)
    mol = Molecule(atoms=atoms, unity='nm')
    mol.set_resname(entry)

    coords = readSection(res, '[ coords ]', '[')
    coords = parseList('fff', coords)
    for i, atom in enumerate(mol.atoms):
        atom.x = coords[i]
        atom.unity = 'nm'
    bonds = []
    bl = readSection(res, '[ bonds ]', '[')
    for line in bl:
        bonds.append(line.split())
    imps = []
    il = readSection(res, '[ impropers ]', '[')
    for line in il:
        imps.append(line.split())
    diheds = []
    dl = readSection(res, '[ dihedrals ]', '[')
    for line in dl:
        diheds.append(line.split())
    rots = []
    rl = readSection(res, '[ rotations ]', '[')
    for line in rl:
        rots.append(line.split())
    rotdic = {}
    for rot in rots:
        key = rot[0]
        rotdic[key] = rot[1:]

    mol.morphes = morphes
    return mol, bonds, imps, diheds, rotdic


def read_mtp(filename='ffoplsaa.mtp'):

    if not hasattr(filename, "read"):
        lst = open(filename).readlines()
    else:
        lst = filename.readlines()
    lst = kickOutComments(lst, ';')

    keyw = ('[ atoms ]', '[ bonds ]', '[ impropers ]',
            '[ dihedrals ]', '[ rotations ]')

    entries = []
    for line in lst:
        if line.startswith('[') and line.strip() not in keyw:
            entries.append(line.strip()[1:-1].strip())
    rdic = {}
    for e in entries:
        rdic[e] = read_mtp_entry(e, filename)
    return rdic



#!/usr/bin/env python
# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

import sys
import os
import shutil
import commands
from glob import glob
from pmx.forcefield import MDP
import argparse


def run_command(func, string):
    s = func.__name__+'(): ' + string
    err_file = func.__name__+'_ERROR.log'
    status, out = commands.getstatusoutput(string)
    if status != 0:
        print >>sys.stderr, 'ERROR in %s ' % func.__name__
        print >>sys.stderr, 'Output written to %s' % err_file
        fp = open(err_file, 'w')
        print >>fp, s
        print >>fp, out
        fp.close()
        sys.exit(1)
    else:
        print "%-90s" % s, ': ok'


def make_dir_tree(skip=4):

    # catch all run?.? dirs
    dirs = glob('run?.?')
    for d in dirs:
        print '\tPreparing run %s' % d
        os.chdir(d)
        os.mkdir('morphes')
        cmd = 'echo 0| trjconv -f traj.trr -s topol.tpr -skip %d -b 2001 -sep -o morphes/frame.gro' % skip
        run_command(make_dir_tree, cmd)
        os.chdir('morphes')

        # put each file into single directory
        file_list = glob('frame*.gro')
        for f in file_list:
            name = f.split('.')[0]
            os.mkdir(name)
            shutil.move(f, name)
        os.chdir('..')
        os.chdir('..')


def prepare_mdp_files(template_mdp, sw_time, sc_alpha, sc_sigma):

    mdp = MDP().read(template_mdp)

    # make 0->1 file
    mdp['free-energy'] = 'yes'
    mdp['init-lambda'] = 0
    # time is in ps -> convert to steps
    nsteps = sw_time*500
    mdp['nsteps'] = int(nsteps)
    # delta lambda = 1./nsteps
    delta_lambda = 1./float(nsteps)
    mdp['delta-lambda'] = delta_lambda
    mdp['sc-alpha'] = sc_alpha
    mdp['sc-sigma'] = sc_sigma

    fp = open('crooks_TI_runA.mdp', 'w')
    print >>fp, mdp
    fp.close()

    # make 1->0 file
    mdp['init-lambda'] = 1
    mdp['delta-lambda'] = -delta_lambda

    fp = open('crooks_TI_runB.mdp', 'w')
    print >>fp, mdp
    fp.close()


def make_run_input_files():

    dirs = glob('run?.?')
    for d in dirs:
        print '\n\tPreparing run input files %s' % d
        mdp_file = None
        if d.split('.')[0][-1] == 'A':  # 0->1
            mdp_file = 'crooks_TI_runA.mdp'
        elif d.split('.')[0][-1] == 'B':  # 1->0
            mdp_file = 'crooks_TI_runB.mdp'
        dir_list = glob(os.path.join(d, 'morphes')+'/frame*')
        for f in dir_list:
            fname = os.path.basename(f)+'.gro'
            gro_file = os.path.join(f, fname)
            cmd = 'grompp -f %s -c %s -o %s/topol.tpr' % (mdp_file, gro_file, f)
            run_command(make_run_input_files, cmd)
            os.system('rm mdout.mdp')


def parse_options():
    parser = argparse.ArgumentParser(description='Script for setting up plain FGTI runs...')

    # TODO: add following options
    # Option( "-box_size", "float", 1.2, "distance from solute to box"),
    # Option( "-conc", "float", 0.15, "ion concentration"),
    # Option( "-vsite", "bool", False, "use virtual sites"),
    # Option( "-fe.dti", "bool", False, "do discrete TI setup"),
    # Option( "-dti_run_time", "float", 10., "Simulation time [ns] at each lambda point"),
    # Option( "-fe.crooks", "bool", False, "do Crooks setup"),
    # Option( "-n_crooks_runs", "int", 1, "setup # crooks runs for each mutation"),
    # Option( "-crooks_run_time", "float", 50., "Simulation time [ns] for crooks equilibrium runs"),
    # Option( "-skip_md_setup", "bool", False, "skip md setup and use -f as starting configuration for free energy runs")
    # FileOption("-crooks_mdp", "r/o",["mdp"],"template", "template run mdp file for crooks equilibrium runs"),
    # FileOption("-dti_mdp", "r/o",["mdp"],"template", "template run mdp file for discrete TI calculations"),
    # FileOption("-min_mdp", "r/o",["mdp"],"em", "template minimization mdp file ( for TI or Crooks )"),
    # FileOption("-lambda_steps", "r/o",["txt"],"lambda_steps", "text file with lambda steps for DTI runs"),

    parser.add_argument('-d',
                        metavar='dir',
                        dest='dir',
                        type=str,
                        help='directory with equlibrated states',
                        default='.')
    parser.add_argument('-mdp',
                        metavar='mdp',
                        dest='mdp',
                        type=str,
                        help='template TI mdp file',
                        default='TI_template.mdp')
    parser.add_argument('-sw_time',
                        metavar='sw_time',
                        dest='sw_time',
                        type=float,
                        help='switching time to use for FGTI runs',
                        default=50.)
    parser.add_argument('-sc_alpha',
                        metavar='sc_alpha',
                        dest='sc_alpha',
                        type=float,
                        help='soft-core alpha',
                        default=0.3)
    parser.add_argument('-sc_sigma',
                        metavar='sc_sigma',
                        dest='sc_sigma',
                        type=float,
                        help='soft-core sigma',
                        default=0.25)
    parser.add_argument('-skip',
                        metavar='skip',
                        dest='skip',
                        type=int,
                        help='skip # frames',
                        default=4)

    args = parser.parse_args()
    return args


def main(args):

    here = os.getcwd()

    run_dir = args.dir
    mdp_file = args.mdp
    sw_time = args.sw_time
    sc_alpha = args.sc_alpha
    sc_sigma = args.sc_sigma

    print '\n\t Preparing FGTI runs in directory..: %s' % run_dir
    print '\t Template mdp file to use............: %s' % mdp_file
    print '\t Switching time to use...............: %8d ps' % int(sw_time)
    print '\t Soft-core alpha to use..............: %8.3f' % sc_alpha
    print '\t Soft-core sigma to use..............: %8.3f' % sc_sigma
    print '\n'


    os.chdir(run_dir)

    print '\t Preparing mdp input files........... '

    prepare_mdp_files(mdp_file, sw_time, sc_alpha, sc_sigma)


    print '\t Preparing directory tree............ '
    make_dir_tree(skip=args.skip)
    make_run_input_files()
    os.chdir(here)
    print '\n\t............... DONE .................\n'


if __name__ == '__main__':
    args = parse_options()
    main(args)



import sys,os
from pmx import *
from pmx.options import *
from pmx.parser import *
from pmx import library
from pmx.mutdb import *
from pmx.geometry import *
from pmx.builder import *

def create_mdp( fname ):
    fp = open(fname,'w')
    fp.write('''
define                  = 
integrator              = steep
emtol                   = 1000.0
nsteps                  = 10000
nstlist                 = 10
cutoff-scheme           = Verlet
rlist                   = 1.2
vdwtype                 = Cut-off
rvdw                    = 1.2
coulombtype             = cutoff
rcoulomb                = 1.2
constraints             = none\n''')
    fp.close()

# identify atoms to be removed and an offset 
def identify_atoms_to_remove( topol, firstResNum=1, lastResNum=2 ):
    removeID = []
    offset = 0
    for a in topol.atoms:
        if a.resnr==firstResNum:
            removeID.append(a.id)
            offset+=1
        if a.resnr==lastResNum:
            removeID.append(a.id)
    return(removeID,offset)

# remove atoms and apply offset
def remove_atoms( topol, removeID ):
    # atoms
    rem = []
    for a in topol.atoms:
        if a.id in removeID:
            rem.append(a)
    for r in rem:
        topol.atoms.remove(r)

    # bonds
    rem = []
    for b in topol.bonds:
        if (b[0].id in removeID) or (b[1].id in removeID):
            rem.append(b)
    for r in rem:
        topol.bonds.remove(r)

    # pairs
    rem = []
    for p in topol.pairs:
        if (p[0].id in removeID) or (p[1].id in removeID):
            rem.append(p)
    for r in rem:
        topol.pairs.remove(r)

    # angles
    rem = []
    for a in topol.angles:
        if (a[0].id in removeID) or (a[1].id in removeID) or (a[2].id in removeID):
            rem.append(a)
    for r in rem:
        topol.angles.remove(r)

    # dihedrals
    rem = []
    for d in topol.dihedrals:
        if (d[0].id in removeID) or (d[1].id in removeID) or (d[2].id in removeID) or (d[3].id in removeID):
            rem.append(d)
    for r in rem:
        topol.dihedrals.remove(r)
  
# renumber atoms in the topology 
def renumber_atoms( topol, offset ):
    # atoms 
    for a in topol.atoms:
        a.id -= offset
        a.cgnr -= offset
        a.resnr -= 1       

# remove residues from topology
def remove_residues( topol ):
    del topol.residues[0]
    del topol.residues[-1]
   
# add angles between the first and last residues
def add_angles( topol, a1, a2, partners ):
    # get 1-2-3 connections
    list13_a1 = []
    list13_a2 = []
    get13( list13_a1, a1, partners, topol )
    get13( list13_a2, a2, partners, topol )

    # add angles
    for a in list13_a1:
        ang = [a[0],a[1],a[2],1]
        ang_inv = [a[2],a[1],a[0],1]
        if (ang in topol.angles) or (ang_inv in topol.angles):
            continue
        else:
            topol.angles.append(ang)
    for a in list13_a2:
        ang = [a[0],a[1],a[2],1]
        ang_inv = [a[2],a[1],a[0],1]
        if (ang in topol.angles) or (ang_inv in topol.angles):
            continue
        else:
            topol.angles.append(ang)
    
# add dihedrals between the first and last residues
def add_dihedrals( topol, a1, a2, partners ):
    # get 1-2-3-4 connections
    list14_a1 = []
    list14_a2 = []
    get14( list14_a1, a1, partners, topol )
    get14( list14_a2, a2, partners, topol )
    # also collect for partners of a1 and a2
    for a in partners[a1]:
        if a!=a2:
            get14( list14_a1, a, partners, topol )
    for a in partners[a2]:
        if a!=a1:
            get14( list14_a2, a, partners, topol )

    # add dihedrals
    for a in list14_a1:
        dih = [a[0],a[1],a[2],a[3],9,'']
        dih_inv = [a[3],a[2],a[1],a[0],9,'']
        if (dih in topol.dihedrals) or (dih_inv in topol.dihedrals):
            continue
        else:
            topol.dihedrals.append(dih)
    for a in list14_a2:
        dih = [a[0],a[1],a[2],a[3],9,'']
        dih_inv = [a[3],a[2],a[1],a[0],9,'']
        if (dih in topol.dihedrals) or (dih_inv in topol.dihedrals):
            continue
        else:
            topol.dihedrals.append(dih)

# add impropers (type 4 for amber)
def add_improper_dihedrals( topol, a1, a2, partners, ff='amber99sb-star-ildn'):
    resnameLast = a2.resname
    resnameFirst = a1.resname
    firstResNum = 1
    lastResNum = len(topol.residues)

    a_N_first = a1
    a_C_last = a2

    # first dihedral: Ca_last - N_first - C_last - O_last
    a_Ca_last = ''
    a_O_last = ''
    for a in topol.atoms:
        if a.resnr==lastResNum and a.name=='CA':
            a_Ca_last = a
        if a.resnr==lastResNum and a.name=='O':
            a_O_last = a
    dih = [a_Ca_last,a_N_first,a_C_last,a_O_last,4,'']
    topol.dihedrals.append(dih)

    # second dihedral: C_last - Ca_first - N_first - H_first
    # (if Pro: C_last - CD_first - N_first - Ca_first  )
    dih = []
    a_Ca_first = ''
    a_H_first = ''
    a_CD_first = ''
    for a in topol.atoms:
        if a.resnr==firstResNum and a.name=='CA':
            a_Ca_first = a
        if a.resnr==firstResNum and a.name=='H':
            a_H_first = a
        if a.resnr==firstResNum and a.name=='CD':
            a_CD_first = a
    if a_N_first.resname!='PRO':
        dih = [a_C_last,a_Ca_first,a_N_first,a_H_first,4,'']
    else:
        dih = [a_C_last,a_CD_first,a_N_first,a_Ca_first,4,'']

    topol.dihedrals.append(dih)

    # for amber-star: N_last - Ca_last - C_last - N_first
    if ('star' in ff) or ('*' in ff):
        if a_C_last.resname!='GLY':
            a_N_last = ''
            for a in topol.atoms:
                if a.resnr==lastResNum and a.name=='N':
                    a_N_last = a
            dih = [a_N_last,a_Ca_last,a_C_last,a_N_first,4,'105.4 0.75 1']
            topol.dihedrals.append(dih)


# add a bond to close the cycle
def add_bond( topol ):
    firstResNum = 1
    lastResNum = len(topol.residues)
    # identify two atoms
    a1 = ''
    a2 = ''
    for a in topol.atoms:
        if a.resnr==firstResNum and a.name=='N':
            a1 = a
        if a.resnr==lastResNum and a.name=='C':
            a2 = a
    # create and add a bond
    newb = [a1,a2,1]
    topol.bonds.insert(0,newb)

    ###### now 1-4 ######
    # create a dictionary of bonded partners for each atom: dict[atom] = [a1,a2,a3,...]
    partners = {}
    for b in topol.bonds:
        if b[0] in partners.keys():
            partners[b[0]].append(b[1])
        else:
            partners[b[0]] = [b[1]]
        if b[1] in partners.keys():
            partners[b[1]].append(b[0])
        else:
            partners[b[1]] = [b[0]]

    # go over bonds and identify 1-4 interactions for a1 and a2 atoms
    pairs = []
    list13 = []
    # atom 1 
    parse14( pairs, list13, a1, partners, topol )
    # atom 2
    parse14( pairs, list13, a2, partners, topol )
    # also partners of a1
    for a in partners[a1]:
        if a!=a2:
            parse14( pairs, list13, a, partners, topol )
    # also partners of a2
    for a in partners[a2]:
        if a!=a1:
            parse14( pairs, list13, a, partners, topol )
    # append
    for p in pairs:
        if [p[0],p[1]] not in list13:
            topol.pairs.append(p)
 
    return(a1,a2,partners)

# parse 1-4 interactions to construct pairs
def parse14( pairs, list13, a, partners, topol ):
    for a12 in partners[a]:
        for a13 in partners[a12]:
            if a13==a:
                continue
            list13.append( [a,a13] )
            list13.append( [a13,a] )
            for a14 in partners[a13]:
                if a12==a14 or a==a14:
                    continue
                p = [a,a14,1]
                pinv = [a14,a,1]
                if (p in pairs) or (pinv in pairs):
                    continue
                if (p in topol.pairs) or (pinv in topol.pairs):
                    continue
                pairs.append(p)

# gets the list of 1-2-3 connections (for angles)
def get13( list13, a, partners, topol ):
    for a12 in partners[a]:
        for a13 in partners[a12]:
            if a13==a:
                continue
            list13.append( [a,a12,a13] )
#            list13.append( [a13,a12,a] )

# gets the list of 1-4 connections (for dihedrals)
def get14( list14, a, partners, topol ):
    for a12 in partners[a]:
        for a13 in partners[a12]:
            if a13==a:
                continue
            for a14 in partners[a13]:
                if a12==a14 or a==a14:
                    continue
                list14.append( [a,a12,a13,a14] )

def main(argv):

   options=[
   Option( "-seq", "string", "PVWLVVV" , "peptide sequence"),
   Option( "-chir", "string", "LDLDLDL" , "peptide sequence"),
   Option( "-ff", "string", "amber99sb-star-ildn" , "for now only amber is supported"),
        ]
    
   files = [
#       FileOption("-seq", "string",[""],"protein.pdb", "input structure of a cyclic peptide"),
       FileOption("-o", "w",["pdb","gro"],"out.pdb", "output structure file"),
       FileOption("-p", "w",["top"],"topol.top", "output topology file"),
       ]
    
   help_text = ('The script builds a cyclic peptide top for gromacs',
                'based on https://github.com/visvaldask/gmx_makecyclictop',
               )

    
   cmdl = Commandline( argv, options = options,
                       fileoptions = files,
                       program_desc = help_text,
                       check_for_existing_files = False )
    
#   iname = cmdl['-f']
   oname = cmdl['-o']
   otopname = cmdl['-p']
   ff = cmdl['-ff']

   # 1. construct the peptide, e.g XYZ
   m = build_chain(cmdl['-seq'],bCyclic=True,chirality=cmdl['-chir'])
   seq = cmdl['-seq']
#   print(seq)

   # 2. add termini, e.g. ZXYZX
   extendedSeq = seq[-1]+seq+seq[0]
#   print(extendedSeq) 
   # take first and last residues
   first_res = m.residues[0].copy()
   last_res = m.residues[-1].copy()
   # append them (there will be overlaps, these residues are just dummy appends)
   m.append(first_res)
   m.insert_residue(0,last_res)
   m.write(oname)

   # 3. pdb2gmx
   cmd = 'gmx pdb2gmx -f out.pdb -ff amber99sb-star-ildn -water tip3p -o out.pdb'
   os.system(cmd)
   # clean after
   cmd = 'rm *# posre.itp'
   os.system(cmd)

   # 4. modify the topology
   topol = Topology( 'topol.top',  version = 'new', assign_types=False )
   removeID,offset = identify_atoms_to_remove( topol, firstResNum=1, lastResNum=len(m.residues) )
   remove_atoms( topol, removeID )   
   renumber_atoms( topol, offset )
   remove_residues( topol )
   a1,a2,partners = add_bond( topol ) # connecting bond and pairs
   # a1 is N in residue 1
   # a2 is C in the last residue
   # partners is a dictionary of bonded partners for each atom: partners[atom] = [a1,a2,a3,...]
   add_angles( topol, a1, a2, partners ) # connecting angles
   add_dihedrals( topol, a1, a2, partners ) # connecting dihedrals
   add_improper_dihedrals( topol, a1, a2, partners, ff=ff) # impropers
 
   # 5. modify the structure: remove first and last residues
   # read the structure from pdb2gmx
   m = Model( 'out.pdb' )
   # get the first and last residues anew
   last_res = m.fetch_residue( len(m.residues) )
   m.remove_residue( last_res )
   first_res = m.fetch_residue( 1 )
   m.remove_residue( first_res )

   # output
   topol.write(otopname)
   m.write(oname)#,bPDBTER=True)

   # 6. minimize
   os.system('mkdir em')
   os.chdir('em')
   # box
   cmd = 'gmx editconf -f ../out.pdb -o box.pdb -bt cubic -box 3 3 3'
   os.system(cmd)
   # grompp
   create_mdp( 'em.mdp' )
   cmd = 'gmx grompp -p ../topol.top -c box.pdb -f em.mdp -o em.tpr'
   os.system(cmd)
   # mdrun
#   mdrun = '/usr/local/gromacs/2021/2021.2-impi2017-fftw337-gcc93-cuda11.1/bin/mdrun_threads'
   mdrun = 'gmx mdrun'
   cmd = '{0} -s em.tpr -c minimized.pdb -v -ntomp 1 -ntmpi 1'.format(mdrun)
   os.system(cmd)
   # clean after
   cmd = 'rm *# md.log mdout.mdp *xtc *trr *edr'
   os.system(cmd)


if __name__=='__main__':
    main(sys.argv)




#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
# pmx is Copyright (C) 2013-2020 by Daniel Seeliger, Vytautas Gapsys
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""Program to build ligand hybrid topology.
"""

import sys
import argparse
from pmx import *
#from pmx.options import *
#from pmx.parser import *
from pmx.model import Model
from pmx.parser import read_and_format
from pmx.utils import get_ff_path, doLog
from pmx.scripts.cli import check_unknown_cmd
from pmx.forcefield import *
from pmx.ndx import *
from pmx.ligand_alchemy import *
import time
import warnings

# =============
# Input Options
# =============
def parse_options():

    parser = argparse.ArgumentParser(description='''
Provided two structures and topologies, build hybrid structure/topology.

''', formatter_class=argparse.RawTextHelpFormatter)

#    exclus = parser.add_mutually_exclusive_group()

    parser.add_argument('-i1',
                        metavar='lig1.pdb',
                        dest='i1',
                        type=str,
                        help='Input ligand structure 1. '
                        'Default is "lig1.pdb"',
                        default='lig1.pdb')
    parser.add_argument('-i2',
                        metavar='lig2.pdb',
                        dest='i2',
                        type=str,
                        help='Input ligand structure 2. '
                        'Default is "lig2.pdb"',
                        default='lig2.pdb')
    parser.add_argument('-itp1',
                        metavar='lig1.itp',
                        dest='itp1',
                        type=str,
                        help='Input ligand topology 1. '
                        'Default is "lig1.itp"',
                        default='lig1.itp')
    parser.add_argument('-itp2',
                        metavar='lig2.itp',
                        dest='itp2',
                        type=str,
                        help='Input ligand topology 2. '
                        'Default is "lig2.itp"',
                        default='lig2.itp')
    parser.add_argument('-pairs',
                        metavar='pairs.dat',
                        dest='pairs',
                        type=str,
                        help='Optional input: atom pair mapping. ')
#                        'Default is "lig2.itp"',
#                        default='lig2.itp')
    parser.add_argument('-n1',
                        metavar='scaffold1.ndx',
                        dest='n1',
                        type=str,
                        help='Optional input: index of atoms to consider for mol1 ')
    parser.add_argument('-n2',
                        metavar='scaffold2.ndx',
                        dest='n2',
                        type=str,
                        help='Optional input: index of atoms to consider for mol2 ')
    parser.add_argument('-oA',
                        metavar='mergedA.pdb',
                        dest='oA',
                        type=str,
                        help='Output: hybrid structure based on the ligand 1. '
                       'Default is "mergedA.pdb"',
                        default='mergedA.pdb')
    parser.add_argument('-oB',
                        metavar='mergedB.pdb',
                        dest='oB',
                        type=str,
                        help='Output: hybrid structure based on the ligand 2. '
                        'Default is "mergedB.pdb"',
                        default='mergedB.pdb')
    parser.add_argument('-oitp',
                        metavar='merged.itp',
                        dest='oitp',
                        type=str,
                        help='Output: hybrid topology. '
                        'Default is "merged.itp"',
                        default='merged.itp')
    parser.add_argument('-offitp',
                        metavar='ffmerged.itp',
                        dest='offitp',
                        type=str,
                        help='Output: atomtypes for hybrid topology. '
                        'Default is "ffmerged.itp"',
                        default='ffmerged.itp')
    parser.add_argument('-log',
                        metavar='hybrid.log',
                        dest='log',
                        type=str,
                        help='Output: log file. '
                        'Default is "hybrid.log"',
                        default='hybrid.log')
    parser.add_argument('--d',
                        metavar='0.05',
                        dest='d',
                        type=float,
                        help='Optional: if -pairs not provided, distance (nm) between atoms to consider them morphable for alignment approach (default 0.05 nm).',
                        default=0.05)
    parser.add_argument('--fit',
                        dest='bFit',
                        help='Fit mol2 onto mol1, only works if pairs.dat is provided',
                        action='store_true')
    parser.add_argument('--split',
                        dest='bSplit',
                        help='split the topology into separate transitions',
                        action='store_true')
    parser.add_argument('--scDUMm',
                        metavar='1.0',
                        dest='scDUMm',
                        type=float,
                        help='scale dummy masses using the counterpart atoms',
                        default=1.0)
    parser.add_argument('--scDUMa',
                        metavar='1.0',
                        dest='scDUMa',
                        type=float,
                        help='scale bonded dummy angle parameters',
                        default=1.0)
    parser.add_argument('--scDUMd',
                        metavar='1.0',
                        dest='scDUMd',
                        type=float,
                        help='scale bonded dummy dihedral parameters',
                        default=1.0)
    parser.add_argument('--deAng',
                        dest='bDeAng',
                        help='decouple angles composed of 1 dummy and 2 non-dummies',
                        action='store_true')

    parser.set_defaults(bFit=False,bSplit=False,bDeAng=False)
    args, unknown = parser.parse_known_args()
    check_unknown_cmd(unknown)

    return args

def entry_point():
    args = parse_options()
    main(args)


def main(args):
    """Run the main script.

    Parameters
    ----------
    args : argparse.Namespace
        The command line arguments
    """

#####################################
    # start timing
    stime = time.time()

#####################################
    # deal with the flags
    bFit = args.bFit
    bSplit = args.bSplit
    bDeAng = args.bDeAng
    d = args.d
    scDUMm = args.scDUMm
    scDUMa = args.scDUMa
    scDUMd = args.scDUMd

#####################################
    # log file
    logfile = open(args.log,'w')

#####################################
    # if pairs.dat is used, ndx will be discarded
    if args.pairs!=None and (args.n1!=None or args.n2!=None):
        doLog(logfile,'pairs.dat is provided, thus discarding index files')
        args.n1 = None
        args.n2 = None
    if args.n1!=None and args.n2==None:
        doLog(logfile,'Only one index file is provided, thus discarding index files')
        args.n1 = None
        args.n2 = None
    if args.n1==None and args.n2!=None:
        doLog(logfile,'Only one index file is provided, thus discarding index files')
        args.n1 = None
        args.n2 = None

#####################################
    # read structures
    doLog(logfile,'Reading ligand 1 from: "%s"' % args.i1)
    m1 = Model().read(args.i1)
    doLog(logfile,'Reading ligand 2 from: "%s"' % args.i2)
    m2 = Model().read(args.i2)

#####################################
    # read topologies
    doLog(logfile,'Reading topology 1 from: "%s"' % args.itp1)
    itp1 = TopolBase(args.itp1)
    doLog(logfile,'Reading topology 2 from: "%s"' % args.itp2)
    itp2 = TopolBase(args.itp2)

#####################################
    # read mapped atom pairs
    if args.pairs!=None:
        doLog(logfile,'Reading file with atom pairs: "%s"' % args.pairs)
        plst = readPairsFile(args.pairs)
    else:
        plst = None

#####################################
    # assign ff parameters
    doLog(logfile,"Assigning forcefield parameters....")
    assignFF(m1,itp1)
    assignFF(m2,itp2)

#####################################
    # read index
    grps = None
    if args.n1!=None and args.n2!=None:
        doLog(logfile,'Reading scaffold index file: "%s"' % args.n1)
        grp1 = IndexFile(args.n1).dic['scaffold']
        doLog(logfile,'Reading scaffold index file: "%s"' % args.n2)
        grp2 = IndexFile(args.n2).dic['scaffold']
        # now we add all atoms with bonds to scaffold atoms
        for b in itp1.bonds:
            if b[0] in grp1.ids and b[1] not in grp1.ids:
                grp1.ids.append(b[1])
                doLog(logfile,'Adding atom %s to scaffold 1' % m1.atoms[b[1]-1].name)
            elif b[1] in grp1.ids and b[0] not in grp1.ids:
                grp1.ids.append(b[0])
                doLog(logfile,'Adding atom %s to scaffold 1' % m1.atoms[b[0]-1].name)
        for b in itp2.bonds:
            if b[0] in grp2.ids and b[1] not in grp2.ids:
                grp2.ids.append(b[1])
                doLog(logfile,'Adding atom %s to scaffold 2' % m2.atoms[b[1]-1].name)
            elif b[1] in grp2.ids and b[0] not in grp2.ids:
                grp2.ids.append(b[0])
                doLog(logfile,'Adding atom %s to scaffold 2' % m2.atoms[b[0]-1].name)
        grps = [grp1.ids, grp2.ids]
    else:
        grps = None

#####################################
    # prepare model copies
    m3 = m1.copy() # m3 will contain all the atoms from m1, but with the coordinates of the matching atoms from m2
    m4 = m2.copy() # need to copy it when fitting

#####################################
    # fitting (optional)

#    if(bFit==True):
    superimposeStructures( args.i1, args.i2, m3, m2, plst, logfile )

#####################################
    # initialize the main object of the LigandHybridTopology class
    doLog(logfile,'Initializing the main object for LigandHybridTopology')
    hybridTop = LigandHybridTopology( m1=m1, m2=m2, m3=m3, m4=m4, itp1=itp1, itp2=itp2, bFit=bFit, scDUMm=scDUMm, scDUMa=scDUMa, scDUMd=scDUMd, bDeAng=bDeAng, pairList=plst, grps=grps, d=d, logfile=logfile )

#####################################
    # the main function: generate hybrid topology
    doLog(logfile,'Starting hybrid topology generation')
    hybridTop._makeHybridTop( )

#####################################
    # write hybrid topology
    doLog(logfile, 'Writing new itp file: "%s"' % args.oitp)
    hybridTop.newitp.write(args.oitp, target_qB = [hybridTop.qB] )
    doLog(logfile, 'Writing dummy forcefield file: "%s"' % args.offitp)
    hybridTop._write_ffitp( args.offitp, m1 )
    # write merged pdb
    if bFit==True:
        m1.write(args.oA)
        m1.write(args.oB)
    else:
        m1.write(args.oA)
        m3.write(args.oB)

#####################################
    # write split topology
    if bSplit==True: # write splitted topology
        hybridTop._write_split_itp( args.oitp )


#*******************************************************************************#
#*******************************************************************************#


    logfile.close()
if __name__ == '__main__':
    entry_point()
